#define _CRT_SECURE_NO_WARNINGS
#include <stdio.h>
#include "common.h"

typedef unsigned short WORD;
typedef unsigned int DWORD;
typedef signed int LONG;
typedef unsigned char BYTE;

#pragma pack(push,2)

// File information header provides general information about the file
typedef struct { 
  WORD    bfType; 
  DWORD   bfSize; 
  WORD    bfReserved1; 
  WORD    bfReserved2; 
  DWORD   bfOffBits; 
} BITMAPFILEHEADER;

// Bitmap information header provides information specific to the image data
typedef struct {
  DWORD  biSize; 
  LONG   biWidth; 
  LONG   biHeight; 
  WORD   biPlanes; 
  WORD   biBitCount; 
  DWORD  biCompression; 
  DWORD  biSizeImage; 
  LONG   biXPelsPerMeter; 
  LONG   biYPelsPerMeter; 
  DWORD  biClrUsed; 
  DWORD  biClrImportant; 
} BITMAPINFOHEADER;

#pragma pack(pop)

void LoadFromBitmap(const char *FileName, int *Width, int *Height, Pixel **Buffer) {
	FILE *f = fopen(FileName, "rb");
	if(f == NULL)
		halt("Cannot open: %s\n", FileName);

	BITMAPFILEHEADER bmf;
	fread(&bmf, sizeof(BITMAPFILEHEADER), 1, f);
	BITMAPINFOHEADER bmi;
	fread(&bmi, sizeof(BITMAPINFOHEADER), 1, f);

	if(bmi.biCompression != 0 || bmi.biPlanes != 1 || bmi.biBitCount != 24)
		halt("Invalid bitmap format: %s\n", FileName);

	fseek(f, bmf.bfOffBits, SEEK_SET);

	int stride = bmi.biWidth*3;
	stride = ((stride+3) & (~3)); // align to DWORD boundary

	int size = bmi.biHeight*stride;
	BYTE *bm = new BYTE[size];
	fread(bm, size, 1, f);
	fclose(f);

	Pixel *dst = new Pixel[bmi.biWidth*bmi.biHeight];

	Pixel *q=dst;
	for(int y=0; y<bmi.biHeight; ++y) {
		int i=(bmi.biHeight-y-1)*stride;
		for(int x=0; x<bmi.biWidth; ++x) {
			int b = bm[i] >> 3;
			int g = bm[i+1] >> 2;
			int r = bm[i+2] >> 3;
			i += 3;
			*q = (Pixel)( (r<<11) | (g<<5) | b );
			++q;
		}
	}

	delete[] bm;

	*Buffer = dst;
	*Width = bmi.biWidth;
	*Height = bmi.biHeight;
}
